#include <adios2.h>
#include <adios2/common/ADIOSMacros.h>
#include <cstdlib>
#include <iostream>
#include <sstream>
#include <stdexcept>
#include <string>
#include <vector>

#include <mpi.h>

struct ADIOS2Params
{
    adios2::ADIOS adios;
    adios2::IO IO;
    adios2::Engine engine;
};

struct Params
{
    ADIOS2Params adios_in;
    ADIOS2Params adios_out;
    int mpi_rank;
    int mpi_size;
    MPI_Comm communicator;
    std::vector<std::unique_ptr<char[]>> writeBuffers;
};

template <typename T>
size_t handle_var(Params &params, std::string const &name)
{
    auto var = params.adios_in.IO.InquireVariable<T>(name);
    if (!var)
    {
        throw std::runtime_error("Failed inquiring variable '" + name + "'");
    }
    auto blocksInfo = params.adios_in.engine.BlocksInfo<T>(
        var, params.adios_in.engine.CurrentStep());
    bool vardefined = false;
    size_t totalExtent = 0;
    for (auto const &info : blocksInfo)
    {
        if (info.BlockID == params.mpi_rank)
        {
            adios2::Variable<T> out_var;
            if (!vardefined)
            {
                out_var = params.adios_out.IO.DefineVariable<T>(
                    name, var.Shape(), info.Start, info.Count);
                vardefined = true;
            }
            else
            {
                out_var = params.adios_out.IO.InquireVariable<T>(name);
            }
            size_t flat_extent = 1;
            for (auto ext : info.Count)
            {
                flat_extent *= ext;
            }
            totalExtent += flat_extent;
            std::unique_ptr<char[]> putFrom(new char[flat_extent * sizeof(T)]{});
            params.adios_out.engine.Put(
                out_var,
                reinterpret_cast<T *>(putFrom.get())); // just write zeroes
            params.writeBuffers.push_back(std::move(putFrom));
        }
    }
    totalExtent *= sizeof(T);
    std::cout << "[" << params.mpi_rank << "] Simulating " << totalExtent
              << "\tbytes for var '" << name << "'" << std::endl;
    return totalExtent;
}

int main(int argsc, char **argsv)
{
    if (argsc != 3)
    {
        std::cout << "Usage: " << argsv[0]
                  << " <sample data location> <out stream file>" << std::endl;
        std::exit(1);
    }

    std::cout << "Initialize MPI" << std::endl;

    MPI_Init_thread(nullptr, nullptr, MPI_THREAD_MULTIPLE, nullptr);

    Params params;
    params.communicator = MPI_COMM_WORLD;
    MPI_Comm_rank(params.communicator, &params.mpi_rank);
    MPI_Comm_size(params.communicator, &params.mpi_size);

    std::cout << "[" << params.mpi_rank << "] Initialized MPI" << std::endl;

    char const *in_file = argsv[1];
    char const *out_file = argsv[2];

    std::cout << "[" << params.mpi_rank << "] Open engines" << std::endl;

    params.adios_in.adios = adios2::ADIOS{params.communicator};
    params.adios_in.IO = params.adios_in.adios.DeclareIO("IO");
    params.adios_in.IO.SetEngine("bp4");
    std::cout << "[" << params.mpi_rank << "] Open BP" << std::endl;
    params.adios_in.engine =
        params.adios_in.IO.Open(in_file, adios2::Mode::Read);
    std::cout << "[" << params.mpi_rank << "] Opened BP" << std::endl;

    params.adios_out.adios = adios2::ADIOS{params.communicator};
    params.adios_out.IO = params.adios_out.adios.DeclareIO("IO");
    params.adios_out.IO.SetEngine("SST");
    params.adios_out.IO.SetParameter("InitialBufferSize", "1Gb");
    params.adios_out.IO.SetParameter("QueueLimit", "1");
    std::cout << "[" << params.mpi_rank << "] Open SST" << std::endl;
    params.adios_out.engine =
        params.adios_out.IO.Open(out_file, adios2::Mode::Write);
    std::cout << "[" << params.mpi_rank << "] Opened SST" << std::endl;

    std::cout << "[" << params.mpi_rank << "] Opened engines" << std::endl;

    std::vector<char> mockHostnameTable;

    while (params.adios_in.engine.BeginStep() == adios2::StepStatus::OK)
    {
        params.adios_out.engine.BeginStep();

        if (params.mpi_rank == 0)
        {
            mockHostnameTable.resize(10 * params.mpi_size);
            auto hosttable = params.adios_out.IO.DefineVariable<char>(
                "/rankTable",
                {size_t(params.mpi_size), 10},
                {0, 0},
                {size_t(params.mpi_size), 10});
            params.adios_out.engine.Put(hosttable, mockHostnameTable.data());
        }
        size_t totalExtent = 0;
        for (auto const &[var, var_meta] :
             params.adios_in.IO.AvailableVariables())
        {
            std::string actualType = var_meta.at("Type");

#define SWITCH_TYPE(ctype, repr)                                               \
    if (#repr == actualType || std::string(#repr) + "_t" == actualType)        \
    {                                                                          \
        totalExtent += handle_var<ctype>(params, var);                         \
    }                                                                          \
    else

            ADIOS2_FOREACH_PRIMITVE_STDTYPE_2ARGS(SWITCH_TYPE)
            {
                throw std::runtime_error("Unknown type: '" + actualType + "'");
            }
#undef SWITCH_TYPE
        }
        std::cout << "[" << params.mpi_rank
                  << "] Closing engines to simulate the sending of "
                  << totalExtent << " bytes" << std::endl;
        params.adios_in.engine.EndStep();
        params.adios_out.engine.EndStep();
        std::cout << "[" << params.mpi_rank << "] Closed engines" << std::endl;
        params.writeBuffers.clear();
    }

    params.adios_out.engine.BeginStep();
    auto var = params.adios_out.IO.DefineVariable<std::string>(
        "adding an additional step");
    std::string messageForYou = "to trigger the queuelimit policy";
    params.adios_out.engine.Put(var, messageForYou);
    std::cout << "Ending the last out step" << std::endl;
    params.adios_out.engine.EndStep();
    std::cout << "Ended the last out step" << std::endl;

    params.adios_in.engine.Close();
    params.adios_out.engine.Close();

    MPI_Barrier(MPI_COMM_WORLD);
}