#include <adios2.h>
#include <adios2/common/ADIOSMacros.h>
#include <cstdlib>
#include <iostream>
#include <memory>
#include <sstream>
#include <stdexcept>
#include <string>
#include <vector>

#include <mpi.h>

struct ADIOS2Params
{
    adios2::ADIOS adios;
    adios2::IO IO;
    adios2::Engine engine;
};

struct OneDimBlock
{
    size_t offset;
    size_t extent;
};

struct Params
{
    ADIOS2Params adios_in;
    int mpi_rank;
    int mpi_size;
    MPI_Comm communicator;
    std::vector<std::unique_ptr<char[]>> loadBuffers;

    OneDimBlock getBlockSlice(size_t total_extent)
    {
        size_t blockSize = total_extent / mpi_size;
        if (total_extent % mpi_size != 0)
        {
            ++blockSize;
        }
        size_t offset = mpi_rank * blockSize;
        size_t extent = blockSize;
        if (offset + extent > total_extent)
        {
            extent = total_extent - offset;
        }
        return OneDimBlock{offset, extent};
    }
};

template <typename T>
void handle_var(Params &params, std::string const &name)
{
    auto var = params.adios_in.IO.InquireVariable<T>(name);
    if (!var)
    {
        throw std::runtime_error("Failed inquiring variable '" + name + "'");
    }

    if (name.find("particlePatches") != std::string::npos)
    {
        /*
         * ParticlePatches are arrays to which every writer rank contributes
         * a low number of entries (in the case of PIConGPU exactly one).
         * (See the decomposition with
         * `bpls simData_00000.bp -De '.*particlePatches.*'`)
         *
         * Loading them to a single reader rank incurs a n->1 communication
         * pattern, where each writer only sends some few bytes of data.
         * On Crusher, this will lead to a hangup at 16 nodes.
         */
#define PARTICLE_PATCHES_ISSUE // this triggers the 16-node issue
#ifdef PARTICLE_PATCHES_ISSUE
        if (params.mpi_rank == 0)
        {
            std::cout << "Loading '" << name << "' fully to rank 0"
                      << std::endl;
            auto extent = var.Shape();
            adios2::Dims offset(extent.size(), 0);
            size_t flat_extent = 1;
            for (auto ext : extent)
            {
                flat_extent *= ext;
            }
            var.SetSelection({std::move(offset), std::move(extent)});
            std::unique_ptr<char[]> loadBuffer(new char[sizeof(T) * flat_extent]);
            params.adios_in.engine.Get(
                var, reinterpret_cast<T *>(loadBuffer.get()));
            params.loadBuffers.push_back(std::move(loadBuffer));
        }
#endif
    }
    else
    {
        /*
         * This is the normal loading of data, which is unproblematic.
         * Each reader loads all blocks written on the same node.
         * (Simulated in simplified way by counting the block IDs)
         */
        auto blocksInfo = params.adios_in.engine.BlocksInfo<T>(
            var, params.adios_in.engine.CurrentStep());
        auto [my_blocks_from, my_blocks_count] =
            params.getBlockSlice(blocksInfo.size());
        for (auto const &info : blocksInfo)
        {
            if (my_blocks_from <= info.BlockID &&
                info.BlockID < my_blocks_from + my_blocks_count)
            {
                auto offset = info.Start;
                auto extent = info.Count;
                size_t flat_extent = 1;
                for (auto ext : extent)
                {
                    flat_extent *= ext;
                }
                var.SetSelection({std::move(offset), std::move(extent)});
                std::unique_ptr<char[]> loadBuffer(
                    new char[sizeof(T) * flat_extent]);
                params.adios_in.engine.Get(
                    var, reinterpret_cast<T *>(loadBuffer.get()));
                params.loadBuffers.push_back(std::move(loadBuffer));
            }
        }
    }
}

int main(int argsc, char **argsv)
{
    if (argsc != 2)
    {
        std::cout << "Usage: " << argsv[0] << " <in stream file>" << std::endl;
        std::exit(1);
    }

    MPI_Init_thread(nullptr, nullptr, MPI_THREAD_MULTIPLE, nullptr);

    Params params;
    params.communicator = MPI_COMM_WORLD;
    MPI_Comm_rank(params.communicator, &params.mpi_rank);
    MPI_Comm_size(params.communicator, &params.mpi_size);

    char const *in_file = argsv[1];
    char const *out_file = argsv[2];

    params.adios_in.adios = adios2::ADIOS{params.communicator};
    params.adios_in.IO = params.adios_in.adios.DeclareIO("IO");
    params.adios_in.IO.SetEngine("sst");
    params.adios_in.IO.SetParameter("OpenTimeoutSecs", "600");
    params.adios_in.engine =
        params.adios_in.IO.Open(in_file, adios2::Mode::Read);

    std::vector<char> mockHostnameTable;

    while (params.adios_in.engine.BeginStep() == adios2::StepStatus::OK)
    {
        /*
         * The ranktable is a small 2D char array written by writer rank 0, and
         * it contains meta information that needs to be distributed to every
         * reader rank.
         * Loading it incurs a communication pattern 1->m, where each reader
         * receives the full char array from rank 0.
         * Doing this on Crusher leads to a hangup at 128 nodes.
         */
#define RANKTABLE_ISSUE
#ifdef RANKTABLE_ISSUE
        auto hosttable = params.adios_in.IO.InquireVariable<char>("/rankTable");
        if (hosttable)
        {
            size_t flat_extent = 1;
            for (auto ext : hosttable.Shape())
            {
                flat_extent *= ext;
            }
            mockHostnameTable.resize(flat_extent);
            hosttable.SetSelection({{0, 0}, hosttable.Shape()});
            params.adios_in.engine.Get(hosttable, mockHostnameTable.data());
        }
        else
        {
            std::cout << "No hosttable in step "
                      << params.adios_in.engine.CurrentStep() << std::endl;
        }
        params.adios_in.engine.PerformGets();
#endif

        // std::cout << "Step " << params.adios_in.engine.CurrentStep() <<
        // ":\n\n";
        for (auto const &[var, var_meta] :
             params.adios_in.IO.AvailableVariables())
        {
            std::string actualType = var_meta.at("Type");

#define SWITCH_TYPE(ctype, repr)                                               \
    if (#repr == actualType || std::string(#repr) + "_t" == actualType)        \
    {                                                                          \
        handle_var<ctype>(params, var);                                        \
    }                                                                          \
    else

            ADIOS2_FOREACH_PRIMITVE_STDTYPE_2ARGS(SWITCH_TYPE)
            if (actualType == "string")
            {
                std::cerr << "Skipping string variable '" << var << "'"
                          << std::endl;
            }
            else
            {
                throw std::runtime_error("Unknown type: '" + actualType + "'");
            }
#undef SWITCH_TYPE
        }
        std::cout << "Closing in engine to load data" << std::endl;
        params.adios_in.engine.EndStep();
        std::cout << "Closed in engine and loaded data" << std::endl;
        params.loadBuffers.clear();
    }

    params.adios_in.engine.Close();

    MPI_Finalize();
}